#     Copyright 2024, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


""" Make a star import, and demonstrate that it has impact on the modules dict.

"""

from __future__ import print_function

from .pyexpat import *  # pylint: disable=unused-wildcard-import,wildcard-import

print("some_package.star_importing, doing the star import")

print("Before", sorted(x for x in dir() if x not in ("__compiled__", "__loader__")))


lala = 1
print("After", sorted(x for x in dir() if x not in ("__compiled__", "__loader__")))

print("Finished")

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
